% from lexicon:
%
% word_entry/2
%   links words in lexicon to lexicon entries (words are strings)
% entry_category/2
%   gives one (possibly complex) category for each lexicon entry
%   categories are given in functional style:
%   rfunc(C1,C2) denotes a CCG functor "C1/C2" which
%     has type C1 and takes an argument C2 to the right
%   lfunc(C1,C2) denotes a CCG functor "C1\C2" which
%     has type C1 and takes an argument C2 to the left
%   (basic categories are strings)

% from input:
%
% word_at/2
%   gives word at position in sentence
%   (word is given as string)
%   (position is given as integer)
% adjacent/2
%   defines adjacency between sentence positions (ordered)

%
% auxiliaries
%

% define positions
position(P) :- word_at(_,P).
minpos(P) :- position(P), P #min [ position(X) = X ] P.
maxpos(P) :- position(P), P #max [ position(X) = X ] P.

% flatten categories used at time point 0, to get all (useful) possible categories for guessing
category(C) :- grid(Px,Py,C).
category(X) :- category(rfunc(X,Y)).
category(Y) :- category(rfunc(X,Y)).
category(X) :- category(lfunc(X,Y)).
category(Y) :- category(lfunc(X,Y)).


% identity
catCanSubstituteFor(Cat,Cat) :- category(Cat).
% subcategory
catCanSubstituteFor(Cat,For) :- subCatOf(Cat,For).
% nested subcategory
% e.g., if S[dcl] isA S, then
% (S[dcl]/C) isA (S/C)
% (S[dcl]\C) isA (S\C)
% (C/S[dcl]) isA (C/S)
% (C\S[dcl]) isA (C\S)
%
% generic formulation:
% if A isA B and C isA D then
% (A/C) isA (B/D)
% (A\C) isA (B\D)
%catCanSubstituteFor(rfunc(ACat,CCat),rfunc(BCat,DCat)) :-
%	category(rfunc(ACat,CCat)),
%	catCanSubstituteFor(ACat,BCat), catCanSubstituteFor(CCat,DCat).
%catCanSubstituteFor(lfunc(ACat,CCat),lfunc(BCat,DCat)) :-
%	category(lfunc(ACat,CCat)),
%	catCanSubstituteFor(ACat,BCat), catCanSubstituteFor(CCat,DCat).

%
% the parser
%

%
% 'declarative' CYK approach (see Figure 4 in "To CNF or not to CNF? An
% Efficient Yet Presentable Version of the CYK Algorithm")
%
% we build a grid where we use only one half triangle:
% x-direction: j = 1..maxpos
% y-direction: i = 1..maxpos
% fields where i = j are categories of words in the sentence
% fields where i < j are unused
% fields where i > j are built from fields left and below
%
% the grounder creates all categories on all grid positions
%
% to get a parse, we guess a path from j=1/i=maxpos to each field where i=j
%

%
% initial situation:
%

% positions i=j have initial categories
%
% each word might have multiple lexicon entries
%
% additionally each position may have multiple tagged
% category entries
%
% all categories are considered at once!
possible_category_for_position(C,P):-position(P),
									word_at(W,P),
									word_entry(W,E),
									entry_category(E,C).
possible_category_for_position(C,P):-position(P),
									category_tag(P,C,Class),
									use_features,
                  use_category_class(Class).
possible_category_for_position(C,P):-position(P),
									category_tag_nofeatures(P,C,Class),
									not use_features,
                  use_category_class(Class).

%%comma and semicolon may be conjunction categor (generic treatment)

possible_category_for_position("conj",P):-possible_category_for_position(",",P).

possible_category_for_position("conj",P):-possible_category_for_position(";",P).


%
% "category convertibility":
% if the tagger returns categories which are convertable into each other for a word,
% e.g., N and NP, then we disallow usage of the less specific one (we lose no parses,
% as all category comparisons are done via catCanSubstituteFor/2)
%
do_no_use_category_for_position(CLessSpecific,P) :-
	 position(P),
   possible_category_for_position(CMoreSpecific,P),
   catCanSubstituteFor(CMoreSpecific,CLessSpecific),
   CMoreSpecific != CLessSpecific.

% use all categories at once! (other approaches guess here, with cyk this is not necessary)
grid_initialcat(P,C) :-
	position(P),
	possible_category_for_position(C,P),
	not do_no_use_category_for_position(C,P).
grid(P,P,C) :-
	grid_initialcat(P,C).

%
% combinator
%

apply_condition_unary(Pj,Pi) :-
	position(Pj;Pi),
	Pj >= Pi.
apply_condition(Pj,Pi,Pleft,Pdown) :-
	position(Pj;Pi;Pleft;Pdown),
	Pj >= Pi,
	Pleft < Pj, % only use categories from the left
	Pleft >= Pi, % do not use unused fields to the left
	Pdown > Pi, % only use categories from downwards
	Pdown <= Pj, % do not use unused fields downwards
	Pdown = Pleft+1. % see CYK algorithm, without this we can skip input words

% mayoccur is a generic predicate for rules that are applied
% mayoccur has arguments (for binary rules):
%   RuleName
%   PositionJTarget (j = left to right)
%   PositionITarget (i = top to bottom)
%   PositionSourceLeft (left source is at PositionSourceLeft/PositionITarget)
%   PositionSourceDown (right source is at PositionJTarget/PositionSourceDown)
%   CreatedCategory category created at PositionJTarget/PositionITarget
%   SourceCategoryLeft (left category used to derive CreatedCategory)
%   SourceCategoryDown (right category used to derive CreatedCategory)
%
% for unary rules we set
%   PositionSourceLeft=PositionJTarget
%   PositionSourceDown=PositionITarget
%   SourceCategoryLeft=SourceCategoryDown (=source category used here) 
grid(Pj,Pi,X) :- mayoccur(_,Pj,Pi,_,_,X,_,_).
grid(Pj,Pi,X) :- mayoccurU(_,Pj,Pi,X,_).

% forward application rule: X/Y Y => X
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = source X position left of Pj
% Pdown = source Y position below Pi
% grid(Pj,Pi,X) :- mayoccur(_,Pj,Pi,_,_,X,_,_).
mayoccur(rule_fwd_appl,Pj,Pi,Pleft,Pdown,X,Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
  catCanSubstituteFor(Left,rfunc(X,Y)),
	catCanSubstituteFor(Right,Y).

% backward application rule: Y X\Y => X
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = source X position left of Pj
% Pdown = source Y position below Pi
mayoccur(rule_bwd_appl,Pj,Pi,Pleft,Pdown,X,Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
	catCanSubstituteFor(Left,Y),
	catCanSubstituteFor(Right,lfunc(X,Y)).

% forward composition rule: X/Y Y/Z => X/Z
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = source X position left of Pj
% Pdown = source Y position below Pi
% grid(Pj,Pi,X) :- mayoccur(_,Pj,Pi,_,_,X,_,_).
mayoccur(rule_fwd_comp,Pj,Pi,Pleft,Pdown,rfunc(X,Z),Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
  catCanSubstituteFor(Left,rfunc(X,Y)),
	catCanSubstituteFor(Right,rfunc(Y,Z)).

% backward composition rule: Y\Z X\Y => X\Z
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = source X position left of Pj
% Pdown = source Y position below Pi
mayoccur(rule_bwd_comp,Pj,Pi,Pleft,Pdown,lfunc(X,Z),Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
	catCanSubstituteFor(Left,lfunc(Y,Z)),
	catCanSubstituteFor(Right,lfunc(X,Y)).

% coordination rule (this is more tricky in CYK as it does not admit ternary rules):
% 1) create conj(X) from conj X
%     ASP allows us to peek whether X will be available later,
%     and doing this would eliminate some grounding,
%     BUT! doing this makes grounding much much slower (11 seconds instead of 0.1)
% 2) create X from X conj(X)
mayoccur(rule_coordination_step1,Pj,Pi,Pleft,Pdown,conj(X),Conj,X) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Conj),
	grid(Pj,Pdown,X),
	conjunction_category(Conj).
mayoccur(rule_coordination_step2,Pj,Pi,Pleft,Pdown,XResult,XLeft,conj(XRight)) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,XLeft),
	grid(Pj,Pdown,conj(XRight)),
  catCanSubstituteFor(XLeft,XResult),
  catCanSubstituteFor(XRight,XResult).

% forward raising rule: X => R/(R\X)
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = Pj
% Pdown = Pi
mayoccurU(rule_fwd_raise,Pj,Pi,rfunc(R,lfunc(R,X)),Xprime) :-
	apply_condition_unary(Pj,Pi),
	grid(Pj,Pi,Xprime),
	catCanSubstituteFor(Xprime,X),
	%X = Xprime,
	limitRaise(X,R).

% backward raising rule: X => R\(R/X)
% Pj/Pi = target position X=Pj, Y=Pi
% Pleft = Pj
% Pdown = Pi
mayoccurU(rule_bwd_raise,Pj,Pi,lfunc(R,rfunc(R,X)),Xprime) :-
	apply_condition_unary(Pj,Pi),
	grid(Pj,Pi,Xprime),
	catCanSubstituteFor(Xprime,X),
	%X = Xprime,
	limitRaise(X,R).

%%%
%%% PUNCTUATION RULES
%%%

% rule_left_comma: , X => X
mayoccur(rule_left_comma,Pj,Pi,Pleft,Pdown,X,",",X) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,","),
	grid(Pj,Pdown,X),
  left_comma(X).

% rule_right_comma: X , => X
mayoccur(rule_right_comma,Pj,Pi,Pleft,Pdown,X,X,",") :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,X),
	grid(Pj,Pdown,","),
  right_comma(X).

% rule_left_colon: : X => X
mayoccur(rule_left_colon,Pj,Pi,Pleft,Pdown,X,":",X) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,":"),
	grid(Pj,Pdown,X),
  left_colon(X).

% rule_right_colon: X : => X
mayoccur(rule_right_colon,Pj,Pi,Pleft,Pdown,X,X,":") :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,X),
	grid(Pj,Pdown,":"),
  right_colon(X).

% rule_left_semicolon: ; X => X
mayoccur(rule_left_semicolon,Pj,Pi,Pleft,Pdown,X,":",X) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,";"),
	grid(Pj,Pdown,X),
  left_colon(X).

% rule_right_semicolon: X ; => X
mayoccur(rule_right_semicolon,Pj,Pi,Pleft,Pdown,X,X,":") :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,X),
	grid(Pj,Pdown,";"),
  right_colon(X).

%
% cross rules
%

% backward cross composition rule: Y/Z X\Y => X/Z
mayoccur(rule_bwd_x_comp,Pj,Pi,Pleft,Pdown,rfunc(X,Z),Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
  catCanSubstituteFor(Left,rfunc(Y,Z)),
  catCanSubstituteFor(Right,lfunc(X,Y)).

% backward cross substitution rule: Y/Z (X\Y)/Z => X/Z
mayoccur(rule_bwd_x_subst,Pj,Pi,Pleft,Pdown,rfunc(X,Z),Left,Right) :-
	apply_condition(Pj,Pi,Pleft,Pdown),
	grid(Pleft,Pi,Left),
	grid(Pj,Pdown,Right),
  catCanSubstituteFor(Left,rfunc(Y,Z)),
  catCanSubstituteFor(Right,rfunc(lfunc(X,Y),Z)).



#hide.
#show word_at/2.
#show adjacent/2.
#show position/1.
#show reachable/3.
#show maxpos/1.
#show minpos/1.
#show occurs_grid/8.
#show occurs_gridU/5.
%#show grid/3.
%#show ban(rule_bwd_appl,J,I,L,D,Rc,Lc).
%#show cat_generated_by/4.
%#show cat_generated_only_by/4.


%
% type raising configuration
%
%
% limitRaise/2 (FromCategory,ToCategory)
%   limits applicability of type raising:
%   to which category raising may apply,
%   and to which category may be raised.
limitRaise("NP","S"). % raise to S
limitRaise("NP",lfunc("S","NP")). % raise to VP = S\NP
limitRaise("NP",rfunc(lfunc("S","NP"),"NP")). % raise to TV = (S\NP)/NP

%%
%% COMBINATOR ONTOLOGY (useful in reification to collapse rules together)
%%
%
% combinator/1 (CombinatorName)
% binary/1 (CombinatorName)
% unary/1 (CombinatorName)
% basic_combinator/1 (CombinatorName)
% application_combinator/1 (CombinatorName)
% composition_combinator/1 (CombinatorName)
%
% combinator properties
%
% combinator_direction/2 (CombinatorName,Direction)
% where Direction is either bwd or fwd


combinator(rule_fwd_appl;rule_bwd_appl).
combinator(rule_coordination_step1;rule_coordination_step2).
combinator(rule_fwd_comp;rule_bwd_comp).
combinator(rule_fwd_raise;rule_bwd_raise).


%
% goal (this requires that there is some parse, we do not know which)
%
finished :-
	maxpos(MaxP),
	minpos(MinP),
	grid(MaxP,MinP,Cat),
  Cat = "S".
  % compatibility with ccg.asp
	%catCanSubstituteFor(Cat,"S").

:- strict_parse, not finished.

%
% getting parse trees from the grid
%

% guess which actions really do occur
0 { occurs_grid(Action,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown) } 1 :-
	mayoccur(Action,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown).
0 { occurs_gridU(Action,Pj,Pi,Result,Src) } 1 :-
	mayoccurU(Action,Pj,Pi,Result,Src).

% at each grid we may have a binary rule application, folowed by a unary rule application

% do not allow multiple binary rules
% -> do not allow multiple rules at one Pj/Pi
%    and using multiple targets different from Pj/Pi
:- position(Pj), position(Pi), 2 #count { occurs_grid(A,Pj,Pi,L,D,R,SL,SR) }.
% do not allow multiple unary rules
% -> do not allow multiple rules at one Pj/Pi
:- position(Pj), position(Pi), 2 #count { occurs_gridU(A,Pj,Pi,R,S) }.
% do not allow unary rule with source different from binary rule target at same Pj/Pi
:- position(Pj), position(Pi),
	occurs_grid(A2,Pj,Pi,Pleft,Pdown,ResultBinary,SLbinary,SRbinary),
	occurs_gridU(A1,Pj,Pi,ResultUnary,SrcUnary),
	ResultBinary != SrcUnary.
% do not allow binary rule using left if unary rule there has different target
:- position(Pj), position(Pi),
	occurs_grid(A2,Pj,Pi,Pleft,Pdown,ResultBinary,SrcLeftBinary,SrcDownBinary),
	occurs_gridU(A1,Pleft,Pi,ResultUnary,SrcUnary),
	ResultUnary != SrcLeftBinary.
% do not allow binary rule using down if unary rule there has different target
:- position(Pj), position(Pi),
	occurs_grid(A2,Pj,Pi,Pleft,Pdown,ResultBinary,SrcLeftBinary,SrcDownBinary),
	occurs_gridU(A1,Pj,Pdown,ResultUnary,SrcUnary),
	ResultUnary != SrcDownBinary.

% normalize only here and only if fact "normalize" is true (allows to disable normalization)
% * we could normalize already for mayoccur
% * BUT doing this is slower
% * AND we must normalize here anyways, as we cannot catch all cases with mayoccur normalization

% ban FA if left category was created by FTR (we can use BA instead)
% schema: [X' Y\X] -> [Y/(Y\X) Y'\X'] --banned--> [Y]
%                              ^-- here Y' and X' is from the FA rule
:- occurs_grid(rule_fwd_appl,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
 	 occurs_gridU(rule_fwd_raise,Pleft,Pi,SrcLeft,Src2),
   normalize.

% ban FA if left category was created by FC (we can use two FAs instead)
% schema: [X/Y Y'/Z Z'] -> [X/Z Z'] --banned--> [X]
%                               ^-- here Z' is from the FA rule
:- occurs_grid(rule_fwd_appl,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	 occurs_grid(rule_fwd_comp,Pleft,Pi,Pleft2,Pdown2,SrcLeft,Src2Left,Src2Down),
   normalize.

% ban BA if right category was created by BTR (we can use FA instead)
% schema: [X/Y Y'] -> [X/Y' X\(X/Y)] --banned--> [X]
%                        ^-- here Y' is from the BA rule
:- occurs_grid(rule_bwd_appl,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	 occurs_gridU(rule_bwd_raise,Pj,Pdown,SrcDown,Src2),
   normalize.

% ban BA if right category was created only by BC (we can use two BAs instead)
% schema: [X' Y'\X Z\Y] -> [X' Z\X] --banned--> [Z]
%                           ^-- here X' is from the BA rule
:- occurs_grid(rule_bwd_appl,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	 occurs_grid(rule_bwd_comp,Pj,Pdown,Pleft2,Pdown2,SrcDown,Src2Left,Src2Right),
   normalize.

% ban BC if left category was created by BC (we can first do BC on the right and then on the left instead)
% schema: [Y'\X Z'\Y V\Z] -> [Z'\X V\Z] --banned--> [V\X]
:- occurs_grid(rule_bwd_comp,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	 occurs_grid(rule_bwd_comp,Pleft,Pi,Pleft2,Pdown2,SrcLeft,Src2Left,Src2Down),
   normalize.

% ban FC if left category was created by FC (we can first do FC on the right and then on the left instead)
% schema: [X/Y Y'/Z Z'/V] -> [X/Z Z'/V] --banned--> [X/V]
:- occurs_grid(rule_fwd_comp,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	 occurs_grid(rule_fwd_comp,Pleft,Pi,Pleft2,Pdown2,SrcLeft,Src2Left,Src2Down),
   normalize.

% ban conjunction 2 if left category was created by FR(X) and right category was
% created by conjunction on FR(X)
% (we can first do conjunction and then FR(X) on the result instead)
% schema:
%    [Y             conj    Y]
%     Pj1/Pi1=YLeft Pj2/Pi2 Pj3/Pi3=YRight
% -> [X/(X\Y)             conj         X/(X\Y)]
%    (created twice by rule_fwd_raise)
%     Pj1/Pi1=SrcLeft Pj2/Pi2=conj Pj3/Pi3=SrcRight
% -> [X/(X\Y)         conj(X/(X\Y))]
%    (created by rule_coordination_step1)
%     Pj1/Pi1=SrcLeft Pj3/Pi2=SrcDown
% --banned--> [X/(X\Y)] % created by rule_coordination_step2
%              Pj3/Pi1=Result
:- occurs_grid(rule_coordination_step2,Pj3,Pi1,Pj1,Pi2,Result,SrcLeft,SrcDown),
	 occurs_gridU(rule_fwd_raise,Pj1,Pi1,rfunc(X,lfunc(X,YLeft)),YLeftOrig),
	 occurs_grid(rule_coordination_step1,Pj3,Pi2,Pj2,Pi3,SrcDown,Conj,SrcRight),
	 occurs_gridU(rule_fwd_raise,Pj3,Pi3,rfunc(X,lfunc(X,YRight)),YRightOrig),
   normalize.
% same for bwd_raise
:- occurs_grid(rule_coordination_step2,Pj3,Pi1,Pj1,Pi2,Result,SrcLeft,SrcDown),
	 occurs_gridU(rule_bwd_raise,Pj1,Pi1,lfunc(X,rfunc(X,YLeft)),YLeftOrig),
	 occurs_grid(rule_coordination_step1,Pj3,Pi2,Pj2,Pi3,SrcDown,Conj,SrcRight),
	 occurs_gridU(rule_bwd_raise,Pj3,Pi3,lfunc(X,rfunc(X,YRight)),YRightOrig),
   normalize.

%
% strict parsing:
% we can mark reachability starting from "S" or a subcategory of "S" at the root
%

% per definition "S" at root is reachable
{ finishCat(Cat) } :-
	strict_parse,
	maxpos(MaxP),
	minpos(MinP),
	grid(MaxP,MinP,Cat),
	catCanSubstituteFor(Cat,"S").
:- strict_parse, not 1 #count { finishCat(Cat) } 1.
reachable(MaxP,MinP,Cat) :-
	strict_parse,
	maxpos(MaxP), minpos(MinP), finishCat(Cat).

%
% non-strict parsing:
% we guess grid positions/categories that shall be marked as reachable
% the fewer grid positions we need to guess to mark all leaves, the better
%

{ guessreachable(X,Y,C) } :-
	not strict_parse,
	grid(X,Y,C).

%
% the following constraints ensure that partial parse trees do not overlap
% this might not be what we really want, perhaps we want no restriction at all
%
% do not guess more than one per X
:- not strict_parse, position(X), 2 { guessreachable(X,Y,C) }.
% do not guess more than one per Y
:- not strict_parse, position(Y), 2 { guessreachable(X,Y,C) }.
% cyk restriction: if Pj/Pi is guessed reachable, then no J/I in the triangle
% to the left and bottom of Pj/Pi must be guessed reachable (or we will get overlapping trees)
% first condition: invalid if J > Pj and I <= Pj (only allow bottom right of the triangle)
:- not strict_parse, position(Pj;Pi;J;I),
	guessreachable(Pj,Pi,C1), guessreachable(J,I,C2), J > Pj, I <= Pj.
% second condition: invalid if J < Pj and J >= Pi (only allow top left of the triangle)
:- not strict_parse, position(Pj;Pi;J;I),
	guessreachable(Pj,Pi,C1), guessreachable(J,I,C2), J < Pj, J >= Pj.

reachable(X,Y,C) :-
	not strict_parse,
	guessreachable(X,Y,C).

minimizationCriterion(X,Y) :-
	not strict_parse,
	guessreachable(X,Y,C).
%#minimize { minimizationCriterion(X,Y) }.

%
% mark reachability to the leaves and enforce that all leaves are reached
% (for non-strict parsing this is always possible as we can simply
% guess all leaves to be reached)
%

% Pj/Pi = position X=Pj, Y=Pi
% Pleft = source X position left of Pj
% Pdown = source Y position below Pi
% binary left
reachable(Pleft,Pi,Cleft) :-
	reachable(Pj,Pi,C),
	occurs_grid(Action,Pj,Pi,Pleft,Pdown,C,Cleft,Cdown).
% binary down
reachable(Pj,Pdown,Cdown) :-
	reachable(Pj,Pi,C),
	occurs_grid(Action,Pj,Pi,Pleft,Pdown,C,Cleft,Cdown).
% unary
reachable(Pj,Pi,Csrc) :-
	reachable(Pj,Pi,C),
	occurs_gridU(Action,Pj,Pi,C,Csrc).

% do not allow rules occuring which are not reachable
:- occurs_grid(A,Pj,Pi,L,D,R,Sl,Sr), not reachable(Pj,Pi,R).
:- occurs_gridU(A,Pj,Pi,R,S), not reachable(Pj,Pi,R).

% enforce that all initial categories at leaves are reachable
reachedLeaf(P) :- grid_initialcat(P,X), reachable(P,P,X).
unreachedLeaf(P) :- position(P), not reachedLeaf(P).
:- unreachedLeaf(P).

%
%
% for debugging only
%
%
%occurs(A,Pj,Pi,R) :- occurs_grid(A,Pj,Pi,L,D,R,Sl,Sr).
%:- not occurs(rule_bwd_app,3,1,"S").
%
%derives(Result,SrcLeft,SrcDown) :-
%	mayoccur(Action,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown).



%
% parser configuration
%

% category ontology:
% subCatOf/2
%subCatOf("S[dcl]","S").
%subCatOf("S[adj]","S").
% subCatOf(lfunc("S","NP"),lfunc("S[dcl]","NP"))
%subCatOf("NP[nb]","NP").
subCatOf("N","NP").
subCatOf("NP",rfunc("S",rfunc("S","NP"))).
%subCatOf("NP",rfunc("NP",lfunc("NP","NP"))).

% make transitives explicit
subCatOf("N",rfunc("S",rfunc("S","NP"))).
%subCatOf("N",rfunc("NP",lfunc("NP","NP"))).

% conjunctions:
% conjunction_category/1
conjunction_category("conj").
conjunction_category("Conj").

% backwards cross substitution licensing

% licensed for Y = "VP-like"

% (see www.ling.umd.edu/~alxndrw/CGReadings/steedman-dowty-1.pdf pg.5)

license_bwd_x_subst(lfunc("S","NP")).

license_bwd_x_subst(rfunc(lfunc("S","NP"),"NP")).

license_bwd_x_subst(rfunc(rfunc(lfunc("S","NP"),"NP"),"NP")).


% punctuation categories
%
% punctuation categories
%
% define "S\S" for "." (full stop)
word_entry(".",fullstop).
entry_category(fullstop,lfunc("S","S")).
% do not allow any other category than "S\S" for "." (full stop)
do_no_use_category_for_position(BadCat,P) :-
	word_at(".",P),
	possible_category_for_position(BadCat,P),
	BadCat != lfunc("S","S").

% see Julia Hockenmaier's thesis pg 34 and 70
% define "(S/S)\NP" for "," (NP extraction left)
word_entry(",",npCommaExtractionL).
entry_category(npCommaExtractionL,lfunc(rfunc("S","S"),"NP")).
% define "(S\S)/NP" for "," (NP extraction right)
word_entry(",",npCommaExtractionR).
entry_category(npCommaExtractionR,rfunc(lfunc("S","S"),"NP")).
% define "((S\NP)\(S\NP))/NP" for "," (NP extraction right)
word_entry(",",npCommaExtractionR).
entry_category(npCommaExtractionR,rfunc(lfunc("S","S"),"NP")).

%% define "(NP\NP)/NP" for ","
%% (NP adjunct to NP, separated by "," as in "Bruce, the brute")
%word_entry(",",commaNPAdjunct).
%entry_category(commaNPAdjunct,rfunc(lfunc("NP","NP"),"NP")).

% define "S[dcl]\S[dcl]" for ","
word_entry(",",emptyComma).
entry_category(emptyComma,lfunc("S[dcl]","S[dcl]")).

% define conjunction for ","
word_entry(",",commaConj).
entry_category(commaConj,"conj").

left_comma("N").
left_comma("NP").
left_comma("S").
left_comma(rfunc("N","N")).
left_comma(lfunc("NP","NP")).
left_comma(lfunc("PP","PP")).
left_comma(rfunc("S","S")).
left_comma(lfunc("S","S")).
left_comma(lfunc("S","NP")).
left_comma(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
left_comma(rfunc(lfunc("S","NP"),lfunc("S","NP"))).
left_comma(lfunc(lfunc(lfunc("S","NP"),lfunc("S","NP")),lfunc(lfunc("S","NP"),lfunc("S","NP")))).

right_comma("N").
right_comma("NP").
right_comma("PP").
right_comma("S").
right_comma(rfunc("N","N")).
right_comma(lfunc("NP","NP")).
right_comma(rfunc("S","S")).
right_comma(lfunc("S","S")).
right_comma(lfunc("S","NP")).
right_comma(rfunc(lfunc("S","NP"),"S")).
right_comma(lfunc(lfunc("S","S"),"NP")).
right_comma(rfunc(lfunc("S","NP"),"NP")).
right_comma(rfunc(lfunc("S","NP"),"PP")).
right_comma(rfunc(lfunc("NP","NP"),lfunc("S","NP"))).
right_comma(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_comma(rfunc(lfunc("S","NP"),lfunc("S","NP"))).

left_colon("N").
left_colon("NP").
left_colon("S").
left_colon(rfunc("N","N")).
left_colon(lfunc("NP","NP")).
left_colon(lfunc("S","NP")).
left_colon(lfunc(lfunc("S","NP"),lfunc("S","NP"))).

right_colon("N").
right_colon("NP").
right_colon("PP").
right_colon("S").
right_colon(lfunc("NP","NP")).
right_colon(rfunc("S","S")).
right_colon(lfunc("S","NP")).
right_colon(rfunc(lfunc("S","NP"),"S")).
right_colon(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_colon(rfunc(lfunc("S","NP"),lfunc("S","NP"))).


left_bracket("N").
left_bracket("NP").
left_bracket("S").
left_bracket(lfunc("NP","NP")).
left_bracket(lfunc(lfunc("S","NP"),lfunc("S","NP"))).

right_bracket("N").
right_bracket("NP").
right_bracket("S").
right_bracket(lfunc("N","N")).
right_bracket(rfunc("N","N")).
right_bracket(lfunc("NP","NP")).
right_bracket(lfunc("S","NP")).
right_bracket(rfunc("S","S")).
right_bracket(lfunc("S","S")).
right_bracket(lfunc(rfunc("N","N"),rfunc("N","N"))).
right_bracket(lfunc(lfunc("S","NP"),lfunc("S","NP"))).
right_bracket(rfunc(lfunc("S","NP"),lfunc("S","NP"))).
