

%%%%
%%%% drawing derivation using IDPDraw (modulo postprocessing of categories)
%%%%

% we draw the word position on x axis
% we draw the time on y axis

#const wordwidth = 9.
#const poswidth = 10.
#const firsttimeheight = 1.
#const timeheight = 2.
#const catheight = 1.
#const actheight = 1.

% object names are tuples
% (first constant is main identifier)
%
% * (aux,Ident)
%   -> auxiliary stuff (captions, ...)
%      (left top coordinate of whatever)
% * (wordpos,Pos) where position(Pos)
%   -> position of surface words
%      (left top coordinate of rectangle)
% * (catpos,Pos,Time) where position(Pos), time(T)
%   -> position of category at timepoint
%      (left top coordinate of rectangle)
% * (actionpos,FirstPos,Time) where position(FirstPos), time(T)
%   -> position of action with first argument FirstPos at timepoint T
%      (left top coordinate of trapezoid)
%      (auxiliary action_width/3: action_width(FirstPos,Time,Width))

%%%%%%%
% words
%%%%%%%

% initial x positions
%
% leftmost word
% -> position X = 0
idpd_xpos_t(wordpos,Pos,X) :-
	X = 0,
	position(Pos),
	{ adjacent(PL,Pos): position(PL) } 0.
%
% words to the right
% -> position X + poswidth
idpd_xpos_t(wordpos,Pos,X+poswidth) :-
	idpd_xpos_t(wordpos,PosLeft,X),
	adjacent(PosLeft,Pos).

% word positions: Y = 0
idpd_ypos_t(wordpos,Pos,Y) :-
	Y = 0,
	position(Pos).

% word text
idpd_text_t(wordpos,P,Word) :- word_at(Word,P).
idpd_polygon_t(4,wordpos,P,0,0,wordwidth,0,wordwidth,1,0,1) :- idpd_text_t(wordpos,P,Text).

%%%%%%%%%%%%%%%%%
% overall polygon (deactivated)
%%%%%%%%%%%%%%%%%

%idpd_xpos_t(aux,overallpoly,0).
%idpd_ypos_t(aux,overallpoly,0).
%idpd_polygon_t(4,aux,overallpoly,0,0,20,0,20,30,30,0).

%%%%%%%%%%%%
% categories
%%%%%%%%%%%%

% x positions exactly where words are
idpd_xpos_t(catpos,P,T,X) :-
	idpd_xpos_t(wordpos,P,X),
	time(T).

% y positions according to time
idpd_ypos_t(catpos,P,0,firsttimeheight) :-
	position(P).
idpd_ypos_t(catpos,P,T+1,Y+timeheight) :-
	idpd_ypos_t(catpos,P,T,Y),
	time(T+1).

% category text
idpd_text_t(catpos,P,T,Text) :- 
	position_category(P,Text,T),
	position_status_open(P,T).
idpd_polygon_t(4,catpos,P,T,0,0,wordwidth,0,wordwidth,catheight,0,catheight) :-
	idpd_text_t(catpos,P,T,Text).



ji2pt(Pj,Pi,P,T) :-
	position(Pj),
	position(Pi),
	minpos(Min),
	maxpos(Max),
	P = Pi,
	T = Pj*2-Pi*2.

raised_gridcat(Pj,Pi,Cat) :-
	occurs_gridU(Action,Pj,Pi,Cat,Src).
nonraised_gridcat(Pj,Pi,Cat) :-
	reachable(Pj,Pi,Cat),
	not raised_gridcat(Pj,Pi,Cat).

% initial times (there are spaces between)

% raised grid categories
position_category_initial(P,Cat,T+1) :-
	raised_gridcat(Pj,Pi,Cat),
	ji2pt(Pj,Pi,P,T).
% nonraised grid categories
position_category_initial(P,Cat,T) :-
	nonraised_gridcat(Pj,Pi,Cat),
	ji2pt(Pj,Pi,P,T).

% binary actions
occurs_initial(Action,L,R,T) :-
	occurs_grid(Action,Pj,Pi,Pleft,Pdown,Result,SrcLeft,SrcDown),
	ji2pt(Pleft,Pi,L,_),
	ji2pt(Pj,Pdown,R,_),
	ji2pt(Pj,Pi,_,T).
% unary actions
occurs_initial(Action,L,dummy,T+1) :-
	occurs_gridU(Action,Pj,Pi,Result,Src),
	ji2pt(Pj,Pi,L,T).

% pushed-together times (tries to minimize spaces)

initialtime(T) :- position_category_initial(_,_,T).
initialtime(T) :- occurs_initial(_,_,_,T).
1 { maptime(T,0..T) } 1 :- initialtime(T).

% enforce unique mapping
:- maptime(T,T1), maptime(T,T2), T1 != T2.
:- maptime(T1,T), maptime(T2,T), T1 != T2.
% enforce ordering as initial times are ordered
:- maptime(T1,T1M), maptime(T2,T2M), T1 > T2, T1M < T2M.
% enforce mapping without empty places in between
maptarget(T) :- maptime(_,T).
% something must be mapped to 0
:- not maptarget(0).
% if something is mapped, the thing below is mapped
:- maptarget(T), T > 0, not maptarget(T-1).

time(T) :- maptarget(T).
					 
position_category(P,Cat,TM) :-
	position_category_initial(P,Cat,T),
	maptime(T,TM).
occurs(Action,L,R,TM) :-
	occurs_initial(Action,L,R,T),
	maptime(T,TM).

position_adjacent(L,R,T) :-
	occurs(Action,L,R,T),
	R != dummy.
position_status_open(P,T) :-
	position_category(P,Cat,T).
position_adjacent(L,R,0) :-
	adjacent(L,R).

%%%%%%%%%
% actions
%%%%%%%%%

% all actions are identified by their first position argument and by T
% actions at T are applied on fluents at T-1 and create fluents at T

% we have several kinds of actions which must be drawn differently:
% actions operating on one position and closing no position (raising)
% actions operating on two positions and closing the right one
% actions operating on three positions and closing left and right ones (coordination)

rule_fwd_raise(P,To,T):-occurs(rule_fwd_raise,P,To,T).
rule_bwd_raise(P,To,T):-occurs(rule_bwd_raise,P,To,T).
rule_fwd_appl(L,R,T):-occurs(rule_fwd_appl,L,R,T).
rule_bwd_appl(L,R,T):-occurs(rule_bwd_appl,L,R,T).
rule_fwd_comp(L,R,T):-occurs(rule_fwd_comp,L,R,T).
rule_bwd_comp(L,R,T):-occurs(rule_bwd_comp,L,R,T).
rule_fwd_x_comp(L,R,T):-occurs(rule_fwd_x_comp,L,R,T).
rule_bwd_x_comp(L,R,T):-occurs(rule_bwd_x_comp,L,R,T).
rule_bwd_x_subst(L,R,T):-occurs(rule_bwd_x_subst,L,R,T).
rule_coordination(L,R,T):-occurs(rule_coordination,L,R,T).
rule_coordination(L,R,T):-occurs(rule_coordination_step1,L,R,T).
rule_coordination(L,R,T):-occurs(rule_coordination_step2,L,R,T).

rule_fwd_appl(L,R,T):-occurs(rule_fwd_appl,L,T), position_adjacent(L,R,T).
rule_bwd_appl(L,R,T):-occurs(rule_bwd_appl,L,T), position_adjacent(L,R,T).
rule_fwd_comp(L,R,T):-occurs(rule_fwd_comp,L,T), position_adjacent(L,R,T).
rule_bwd_comp(L,R,T):-occurs(rule_bwd_comp,L,T), position_adjacent(L,R,T).
rule_fwd_x_comp(L,R,T):-occurs(rule_fwd_x_comp,L,T), position_adjacent(L,R,T).
rule_bwd_x_comp(L,R,T):-occurs(rule_bwd_x_comp,L,T), position_adjacent(L,R,T).
rule_bwd_x_subst(L,R,T):-occurs(rule_bwd_x_subst,L,T), position_adjacent(L,R,T).
rule_coordination(L,R,T):-occurs(rule_coordination,L,T), position_adjacent(L,M,T),position_adjacent(M,R,T).

% one_position_action(Position,Time,Descriptiontext)
one_position_action(P,T+1,ftr) :- rule_fwd_raise(P,To,T).
one_position_action(P,T+1,btr) :- rule_bwd_raise(P,To,T).

% two_position_action(LPosition,Time,RPosition,Descriptiontext)
two_position_action(L,T+1,R,fa) :- rule_fwd_appl(L,R,T).
two_position_action(L,T+1,R,ba) :- rule_bwd_appl(L,R,T).
two_position_action(L,T+1,R,fc) :- rule_fwd_comp(L,R,T).
two_position_action(L,T+1,R,bc) :- rule_bwd_comp(L,R,T).
two_position_action(L,T+1,R,fxc) :- rule_fwd_x_comp(L,R,T).
two_position_action(L,T+1,R,bxc) :- rule_bwd_x_comp(L,R,T).
two_position_action(L,T+1,R,bxs) :- rule_bwd_x_subst(L,R,T).

two_position_action(L,T+1,R,commal):-occurs(rule_left_comma,L,R,T).
two_position_action(L,T+1,R,commar):-occurs(rule_right_comma,L,R,T).
two_position_action(L,T+1,R,colonl):-occurs(rule_left_colon,L,R,T).
two_position_action(L,T+1,R,colonr):-occurs(rule_right_colon,L,R,T).
two_position_action(L,T+1,R,scolonl):-occurs(rule_left_semicolon,L,R,T).
two_position_action(L,T+1,R,scolonr):-occurs(rule_right_semicolon,L,R,T).

two_position_action(L,T+1,R,fa_rf) :- occurs(rule_fwd_appl_rf,L,R,T).
two_position_action(L,T+1,R,fa_rb) :- occurs(rule_fwd_appl_rb,L,R,T).
two_position_action(L,T+1,R,ba_lf) :- occurs(rule_bwd_appl_lf,L,R,T).
two_position_action(L,T+1,R,ba_lb) :- occurs(rule_bwd_appl_lb,L,R,T).
two_position_action(L,T+1,R,fc_lf) :- occurs(rule_fwd_comp_lf,L,R,T).
two_position_action(L,T+1,R,fc_rf) :- occurs(rule_fwd_comp_rf,L,R,T).
two_position_action(L,T+1,R,bc_lb) :- occurs(rule_bwd_comp_lb,L,R,T).
two_position_action(L,T+1,R,bc_rb) :- occurs(rule_bwd_comp_rb,L,R,T).
two_position_action(L,T+1,R,fc_lfrf) :- occurs(rule_fwd_comp_lfrf,L,R,T).
two_position_action(L,T+1,R,bc_lbrb) :- occurs(rule_bwd_comp_lbrb,L,R,T).

% three_position_action(LPosition,Time,MPosition,RPosition,Descriptiontext)
three_position_action(L,T+1,M,R,conj) :- rule_coordination(L,R,T),
	position_adjacent(L,M,T).





% action descriptions are identified by leftmost positions
action_desc(L,T,Desc) :- one_position_action(L,T,Desc).
action_desc(L,T,Desc) :- two_position_action(L,T,R,Desc).
action_desc(L,T,Desc) :- three_position_action(L,T,M,R,Desc).

% one position actions have a width of the width of the position - 2
action_width(P,T,wordwidth-2) :- one_position_action(P,T,Desc).
% one position actions are positioned to the center of the position
% -> given a width of position with - 2, they are positioned at position+1
idpd_xpos_t(actionpos,P,T,X+1) :-
	one_position_action(P,T,Desc),
	idpd_xpos_t(catpos,P,T-1,X),
	time(T-1).

% two position actions are positioned to the center of the left position
idpd_xpos_t(actionpos,L,T,X2) :-
	two_position_action(L,T,R,Desc),
	idpd_xpos_t(catpos,L,T-1,X1),
	time(T-1),
	X2 = (X1+wordwidth/2).
% two position actions have a width of the distance between their two positions
action_width(L,T,W) :- two_position_action(L,T,R,Desc),
	idpd_xpos_t(catpos,L,T-1,X1),
	idpd_xpos_t(catpos,R,T-1,X2),
	W = X2 - X1.

% three position actions are positioned to the center of the left position
idpd_xpos_t(actionpos,L,T,X2) :-
	three_position_action(L,T,M,R,Desc),
	idpd_xpos_t(catpos,L,T-1,X1),
	time(T-1),
	X2 = (X1+wordwidth/2).
% three position actions have a width of the distance between their left and right positions
action_width(L,T,W) :- three_position_action(L,T,M,R,Desc),
	idpd_xpos_t(catpos,L,T-1,X1),
	idpd_xpos_t(catpos,R,T-1,X2),
	W = X2 - X1.

% actions are above category of same time
idpd_ypos_t(actionpos,P,T,Y-catheight) :-
	action_desc(P,T,Desc),
	idpd_ypos_t(catpos,P,T-1,Y).

% text for action
idpd_text_t(actionpos,P,T,Desc) :-
	action_desc(P,T,Desc).
% polygon for action
idpd_polygon_t(4,actionpos,P,T,0,0,Width,0,Width-1,1,1,1) :-
	action_desc(P,T,Desc),
	action_width(P,T,Width).

%#hide.
%#show idpd_xpos_t/3.
%#show idpd_xpos_t/4.
%#show idpd_ypos_t/3.
%#show idpd_ypos_t/4.
%#show idpd_text_t/3.
%#show idpd_text_t/4.
%#show idpd_color_t/5.
%#show idpd_color_t/6.
%#show idpd_pencolor_t/5.
%#show idpd_pencolor_t/6.
%#show idpd_polygon_t/11.
%#show idpd_polygon_t/12.

% vim:syn=prolog:


